package core.notevault.sync;
import android.content.Context;
import android.util.Log;
import androidx.annotation.NonNull;
import androidx.work.Worker;
import androidx.work.WorkerParameters;
import core.notevault.data.Concert;
import core.notevault.data.MusicDatabase;
import core.notevault.data.MusicNoteDAO;
import core.notevault.data.sync.SyncAction;
import core.notevault.data.sync.SyncObject;
import core.notevault.data.sync.SyncResponse;
import core.notevault.data.sync.SyncStatus;
import core.notevault.sync.synchronisation.*;
import retrofit2.Call;
import retrofit2.Callback;
import retrofit2.Response;

import java.io.IOException;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.stream.Collectors;

public class SyncWorker extends Worker{
    private SyncService syncService;
    private MusicNoteDAO database;
    public SyncWorker(@NonNull Context context, @NonNull WorkerParameters workerParams) {
        super(context, workerParams);
        this.syncService = ApiClient.getRetrofitInstance(context).create(SyncService.class);
        MusicDatabase musicDatabase = MusicDatabase.getDatabase(context);
        database = musicDatabase.musicNoteDao();
    }

    @NonNull
    @Override
    public Result doWork() {
        SyncRequest concertSync = buildConcertSyncRequest();

        final CountDownLatch latch = new CountDownLatch(1);
        final Result[] result = {Result.failure()};

        syncWithServer(concertSync, new Callback<SyncResponseModel>() {
            @Override
            public void onResponse(Call<SyncResponseModel> call, Response<SyncResponseModel> response) {
                if(response.isSuccessful()) {
                    result[0] = Result.success();
                    Log.d("SyncWorker", "Sync finished, process now");
                    processConcertSynResponse(response.body());
                } else {
                    Log.d("SyncWorker", "Sync failed");
                    if (response.errorBody() != null) {
                        try {
                            String errorBody = response.errorBody().string();
                            Log.d("SyncWorker", "Error Body: " + errorBody);
                        } catch (IOException e) {
                            e.printStackTrace();
                        }
                    }
                    result[0] = Result.failure();
                }
                latch.countDown();
            }

            @Override
            public void onFailure(Call<SyncResponseModel> call, Throwable throwable) {
                result[0] = Result.failure();
                Log.d("SyncWorker", "Sync failure");
                latch.countDown();
            }
        });

        try {
            latch.await();
        } catch (InterruptedException e) {
//            throw new RuntimeException(e);
        }
        return result[0];
    }

    private SyncRequest buildConcertSyncRequest() {
        List<Concert> createdConcerts = database.getConcertsWithSyncStatus(SyncStatus.CREATED);
        List<Concert> deletedConcerts = database.getConcertsWithSyncStatus(SyncStatus.DELETED);
        List<Concert> modifiedConcerts = database.getConcertsWithSyncStatus(SyncStatus.MODIFIED);


        List<Integer> deviceIDsCreatedConcerts = createdConcerts.stream().map(Concert::getId).collect(Collectors.toList());
        List<SyncChangeRequest> modifiedSyncRequests = modifiedConcerts.stream().map(concert ->
                new SyncChangeRequest(concert.getServerSpecificConcertID(), concert.getLastModified())).collect(Collectors.toList());
        List<SyncChangeRequest> deletedSyncRequests = deletedConcerts.stream().map(concert ->
                new SyncChangeRequest(concert.getServerSpecificConcertID(), concert.getLastModified())).collect(Collectors.toList());

        Log.d("SyncWorker", "Build SyncRequest");
        // Datenbankabfrage: Nur IDs mit Zustand "Modified" oder "Deleted" abrufen
        // Dies ist abhängig von der spezifischen Implementierung der Datenbank
        return new SyncRequest(database.getLatestSyncTimestamp(), deletedSyncRequests, modifiedSyncRequests, deviceIDsCreatedConcerts);
    }

    private void syncWithServer(SyncRequest syncRequest, Callback<SyncResponseModel> callback) {
        syncService.syncConcerts(syncRequest).enqueue(callback);
    }

    private void processConcertSynResponse(SyncResponseModel syncResponse) {
        Log.d("SyncWorker", "Process SyncResponse");
        //Assign local Concert entries the provided Server ID
        List<SyncResponse> syncResponses = new ArrayList<>();
        for(SyncCreateResponse createResponse : syncResponse.getCreateResonses()) {
            database.updateConcertServerUUID(createResponse.getServerObjectUUID(), createResponse.getDeviceSpecificObjectID());
            syncResponses.add(new SyncResponse(createResponse.getServerObjectUUID(), SyncAction.TOBEUPLOADED, SyncObject.CONCERT));
        }

        //Delete Concerts that the server has marked as deleted
        database.deleteConcerts(syncResponse.getToBeDeleted());

        //Persist SyncResponse

        for(String toBeDownloaded : syncResponse.getToBeDownloaded()) {
            syncResponses.add(new SyncResponse(toBeDownloaded, SyncAction.TOBEDOWNLOADED, SyncObject.CONCERT));
        }

        for(String toBeUploaded : syncResponse.getToBeUploaded()) {
            syncResponses.add(new SyncResponse(toBeUploaded, SyncAction.TOBEUPLOADED, SyncObject.CONCERT));
        }
        database.insertSyncResponses(syncResponses);
    }
}
